#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <string.h>
#include <limits.h>

int equal_node (struct stat st1, struct stat st2) {
  return st1.st_dev == st2.st_dev && st1.st_ino == st2.st_ino;
}

int is_root(char *dir) {
  struct stat current;
  struct stat parent;
  char parent_dir[PATH_MAX];
  if ( -1 == stat(dir, &current) ) {
    perror(dir);
    return 0;
  }
  strlcpy(parent_dir, dir, PATH_MAX);
  strlcat(parent_dir, "/..", PATH_MAX);
  if ( -1 == stat(parent_dir, &parent) ) {
    perror(parent_dir);
    return 0;
  }
  return equal_node(current, parent);
}

char * dir_find (char * dir_name, struct stat file_stat) {
  DIR * dir;
  struct dirent * entry;
  struct stat entry_stat;
  char entry_name[PATH_MAX];

  if ( NULL == (dir = opendir(dir_name)) ) {
    perror("dir_find: opendir");
    return NULL;
  }
  while ( NULL != (entry = readdir(dir)) ) {
    strlcpy(entry_name, dir_name, PATH_MAX);
    strlcat(entry_name, "/", PATH_MAX);
    strlcat(entry_name, entry->d_name, PATH_MAX);

    if ( -1 == (lstat(entry_name, &entry_stat)) ) {
      perror("dir_find: lstat");
      closedir(dir);
      return NULL;
    }
    if ( equal_node (file_stat, entry_stat) ) {
      char * res;
      closedir(dir);
      if ( NULL == (res = (char *) calloc(PATH_MAX, sizeof(char))) ) {
  	perror("dir_find: calloc");
  	return NULL;
      }
      strlcpy(res, entry->d_name, PATH_MAX);
      return res;
    }
  }
  closedir(dir);
  return NULL;
}


char * mon_getcwd_aux (char * buf, size_t size,
		       char * dir, size_t dir_size)
{
  struct stat current_st;
  char * path;
  char * dir_name;

  if ( is_root(dir) ) {
    strlcpy(buf, "/", size);
    return buf;
  }

  if ( -1 == stat(dir, &current_st) ) {
    perror("mon_getcwd_aux: stat");
    return NULL;
  }
  strlcat(dir, "/..", dir_size);
  if ( NULL ==  (dir_name = dir_find (dir, current_st)) ) {
    return NULL;
  }
  if ( NULL == (path = mon_getcwd_aux(buf, size, dir, dir_size)) ) {
    return NULL;
  }
  strlcat(path, dir_name, size);
  strlcat(path, "/", size);
  free(dir_name);
  return path;
}

char * mon_getcwd (char * buf, size_t size) {
  char dir[PATH_MAX];
  if ( NULL == buf) {
    size = PATH_MAX;
    if ( NULL == (buf = (char *) calloc(PATH_MAX, sizeof(char))) ) {
      perror("mon_getcwd: calloc");
      return NULL;
    }
  }
  strlcpy(dir, ".", PATH_MAX);
  return mon_getcwd_aux(buf, size, dir, PATH_MAX);
}


int main(int argc, char **argv) {
  char * path;

  if ( NULL == (path = (char *) calloc(PATH_MAX, sizeof(char))) ) {
    perror("main: calloc");
    exit(1);
  }
  if ( NULL == mon_getcwd(path, PATH_MAX) ) {
    perror("main: getcwd");
    exit(1);
  }
  printf("%s\n", path);
  exit(0);
}
